//
//  MyController.m
//  TestSignals
//
//  Created by Cameron Hayne on Sun Nov 16 2003.
//  Copyright (c) 2005 Hayne of Tintagel. All rights reserved.
//

#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#import "MyController.h"

@interface MyController (privateMethods)
- (void)handleSIGHUP;
- (void)handleSIGINT;
- (void)handleSIGQUIT;
- (void)handleSIGABRT;
- (void)handleSIGKILL;
- (void)handleSIGTERM;
- (void)handleSIGSTOP;
- (void)handleSIGUSR1;
- (void)handleSIGUSR2;
- (void)handleOtherSignals:(int)signalNum;
- (void)startSignalWatcherTimer;
- (void)startProgressTimer;
- (void)appendMessage;
@end

static int           myLastSignal = 0;

// Note from Greg Parker (of Apple) about signal handlers:
// (posted to the Cocoa-Dev mailing list, November 2004)
// ---------------------------------------------------------------------------
// In general, there's very little you can do safely from inside a signal 
// handler, because the signal handler interrupts the handling thread at 
// an unpredictable point. For example, the handling thread could have been 
// in the middle of malloc() and currently hold the malloc lock. If the 
// signal handler also calls malloc(), it will hang when it tries to grab 
// the malloc lock.
//
// In Cocoa, objc_msgSend() may call malloc() or grab other locks from 
// the Objective-C runtime itself. Usually it doesn't, and your signal 
// handler survives, but not always. 
//
// Among the few signal-safe functions and operations are: 
// * set a global variable that some other thread looks at
// * call _exit(2) or _Exit(3) (but not exit(3)!)
// * call write(1)
// * call mach_msg_send()
//
// One common solution is to write() to a file descriptor that some 
// other thread is blocked in read(), and have that other thread do 
// the real work. Similarly, you can use mach_msg_send() to wake an 
// NSMachPort or CFMachPort in a runloop.
// ---------------------------------------------------------------------------

void catcher(int sigraised)
{
    // Set a global variable to indicate that a signal was caught.
    // This variable is read in the 'showSignalInfo' method
    // that is invoked via a timer.
    // (See the above note from Greg Parker about what is safe to do
    // in a signal handler)
    
    myLastSignal = sigraised;
}

void setupSignalCatchers()
{
    (void) signal(SIGHUP,  catcher);
    (void) signal(SIGINT,  catcher);
    (void) signal(SIGQUIT, catcher);
    (void) signal(SIGABRT, catcher);
    (void) signal(SIGKILL, catcher);
    (void) signal(SIGTERM, catcher);
    (void) signal(SIGSTOP, catcher);
    (void) signal(SIGUSR1, catcher);
    (void) signal(SIGUSR2, catcher);
}


@implementation MyController

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
    myCount = 0;
    [countTxt setIntValue: myCount];
    myPid = getpid();
    [pidTxt setIntValue: myPid];

    setupSignalCatchers();
    [self startSignalWatcherTimer];
    [self startProgressTimer];
}

// closing the window quits the app
- (BOOL)applicationShouldTerminateAfterLastWindowClosed:(NSApplication *)theApplication
{
    return YES;
}

// this method is invoked via a timer
- (void)showProgress:(NSTimer *)aTimer
{
    ++myCount;
    [countTxt setIntValue: myCount];
}

- (void)startProgressTimer
{
    [NSTimer scheduledTimerWithTimeInterval: 1.0
                                     target: self
                                   selector: @selector(showProgress:)
                                   userInfo: nil
                                    repeats: YES];
}

// this method is invoked via a timer 
// Note that this is used as a simple method of side-stepping the constraints
// of what is safe to do in a signal handler. (See above note)
// In a real application, you would probably handle signals using a separate
// thread as outlined in the above note.
- (void)showSignalInfo:(NSTimer *)aTimer
{
    if (myLastSignal <= 0)
    {
        // no signal has been received
        return;
    }
    
    switch (myLastSignal)
    {
        case SIGHUP:  // 1
            [self handleSIGHUP];
            break;
            
        case SIGINT:  // 2
            [self handleSIGINT];
            break;
            
        case SIGQUIT: // 3
            [self handleSIGQUIT];
            break;
            
        case SIGABRT: // 6
            [self handleSIGABRT];
            break;
            
        case SIGKILL: // 9
            [self handleSIGKILL];
            break;
            
        case SIGTERM: // 15
            [self handleSIGTERM];
            break;
            
        case SIGSTOP: // 17
            [self handleSIGSTOP];
            break;
            
        case SIGUSR1: // 30
            [self handleSIGUSR1];
            break;
            
        case SIGUSR2: // 31
            [self handleSIGUSR2];
            break;
            
        default:
            [self handleOtherSignals: myLastSignal];
            break;
    }
    
    myLastSignal = 0;
}

- (void)startSignalWatcherTimer
{
    [NSTimer scheduledTimerWithTimeInterval: 0.3
                                     target: self
                                   selector: @selector(showSignalInfo:)
                                   userInfo: nil
                                    repeats: YES];
}

- (void)appendMessage:(NSString *)msg
{
    [messagesTxt setString:
                    [[messagesTxt string] stringByAppendingString: msg]];
    [messagesTxt scrollRangeToVisible:
                    NSMakeRange([[messagesTxt string] length], 0)];
}

- (void)handleSIGHUP
{
    [self appendMessage: @"Received HUP signal (1).\n"];
    [self appendMessage: @"-> resetting count to zero.\n"];
    myCount = 0;
    [countTxt setIntValue: myCount];
}

- (void)handleSIGINT
{
    [self appendMessage: @"Received INT signal (2).\n"];
    [self appendMessage: @"-> do it again to get the default action.\n"];
    (void) signal(SIGINT,  SIG_DFL);
}

- (void)handleSIGQUIT
{
    [self appendMessage: @"Received QUIT signal (3).\n"];
    [self appendMessage: @"-> but I'm not going to do anything about it!\n"];
}

- (void)handleSIGABRT
{
    [self appendMessage: @"Received ABRT signal (6).\n"];
    [self appendMessage: @"-> but I'm not going to do anything about it!\n"];
}

- (void)handleSIGKILL
{
    // this should never happen - KILL cannot be caught!
    [self appendMessage: @"Received KILL signal (9).\n"];
    [self appendMessage: @"-> why are we here?\n"];
}

- (void)handleSIGTERM
{
    [self appendMessage: @"Received TERM signal (15).\n"];
    [self appendMessage: @"-> do it again to get the default action.\n"];
    (void) signal(SIGTERM,  SIG_DFL);
}

- (void)handleSIGSTOP
{
    // this should never happen - STOP cannot be caught!
    [self appendMessage: @"Received STOP signal (17).\n"];
    [self appendMessage: @"-> why are we here?\n"];
}

- (void)handleSIGUSR1
{
    [self appendMessage: @"Received USR1 signal (30).\n"];
    [self appendMessage: @"-> changing colour to green.\n"];
    NSColor *green = [NSColor greenColor];
    [countTxt setBackgroundColor: green];
}

- (void)handleSIGUSR2
{
    [self appendMessage: @"Received USR2 signal (31).\n"];
    [self appendMessage: @"-> changing colour to blue.\n"];
    NSColor *blue = [NSColor blueColor];
    [countTxt setBackgroundColor: blue];
}

- (void)handleOtherSignals:(int)signalNum
{
    [self appendMessage: [NSString stringWithFormat:
                            @"Received signal (%d).\n", signalNum]];
}

// this is the action for the pull-down menu
// it sends a signal to the current process
- (IBAction)signalAction:(id)sender
{
    int signalNum = [[sender selectedItem] tag];
    kill (myPid, signalNum);
}

@end
